#!/usr/bin/python

from charmhelpers.core.hookenv import (
    log,
    DEBUG,
    ERROR,
    config,
    status_get,
    status_set,
    relation_ids,
    related_units,
    relation_get,
    storage_get,
    Hooks,
    UnregisteredHookError
)
from charmhelpers.fetch import (
    apt_update,
    apt_upgrade,
)
from charmhelpers.core.host import (
    service
)
from quobyte_helpers import (
    install_quobyte,
    is_empty,
    call,
    create_filesystem,
    is_mounted,
    mount_device,
    is_quobyte,
    is_installed,
    do_update_replica_hosts
)
import json
import sys

hooks = Hooks()


def service_restart():
    config_data = config()

    # ensure software is installed
    if not is_installed():
        result = install_quobyte()
        if result is False:
            # could not install (missing repr_id?), propagate exitcode
            return 0

    # ensure mount-points are prepared as quobyte-data devices
    mount_points = config_data['mount_points']
    if mount_points is not None:
        for mount_point in mount_points:
            if not is_quobyte(mount_point):
                success = do_device_preparation(mount_point)
                log('do_device_preparation yields a {}'.format(success), DEBUG)
            else:
                log('{} is a quobyte device. empty: {}'
                    .format(mount_point, is_empty(mount_point)), DEBUG)
                pass
    else:
        msg = 'no storage devices given, waiting for storage relation'
        log(msg)
        status_set('waiting', msg)
        return 0

    # ensure registry-hosts are given and  /etc/quobyte/host.cfg is updated
    result = []
    for rel_id in relation_ids('quobyte-registry'):
        for unit in related_units(rel_id):
            replicas = relation_get('replicas', unit, rel_id)
            if replicas is not None:
                log('grep replicas is {} {}'.format(replicas, type(replicas)))
                result = json.loads(replicas)
                log('grep list is {} {}'.format(result, type(result)))
                break
    if len(result) is 0:
        msg = 'no quobyte-registry-relation given, waiting for registry'
        log(msg)
        status_set('waiting', msg)
        return 0

    success, hosts = do_update_replica_hosts(result)
    if success:
        log('/etc/quobyte/host.cfg updated: registry={}'.format(hosts))
    else:
        log('cannot update /etc/quobyte/host.cfg', ERROR)
        status_set('blocked', 'cannot update /etc/quobyte/host.cfg')

    status_set('active', 'Quobyte-Metadata Service running')
    return service('restart', 'quobyte-metadata')


def service_stop():
    if status_get()[0] == 'active':
        status_set('maintenance', 'Quobyte-Metadata Service stopped')
    return service('stop', 'quobyte-metadata')


def do_device_preparation(mount_point):
    config_data = config()
    overwrite = config_data['overwrite']
    if is_empty(mount_point) or (overwrite is True and not is_quobyte(mount_point)):
        log('{} is empty, creating Quobyte Metadata Device'.format(mount_point))
        cmd = ['qmkdev', '-t', 'METADATA', mount_point]
        success, output = call(cmd)
        if success:
            log(output)
        else:
            log(output, ERROR)
        return success
    else:
        log('ERROR: {} not empty, do not initialize Quobyte Data'.format(mount_point))
        return 1


# -------------------------------------------


@hooks.hook('start')
def start():
    return service_restart()


@hooks.hook('stop')
def stop():
    return service_stop()


@hooks.hook('config-changed')
def config_changed():
    return install_quobyte()


@hooks.hook('upgrade-charm')
def upgrade_charm():
    service_stop()
    apt_update(fatal=True)
    apt_upgrade(fatal=True)
    return service_restart()


@hooks.hook('install')
def install():
    return install_quobyte()


# -------------------------------------------

@hooks.hook('metadata-storage-storage-attached')
def metadata_storage_storage_attached():
    log('metadata-storage-storage-attached hook called', DEBUG)

    config_data = config()
    overwrite = config_data['overwrite']
    if 'mount_points' not in config_data:
        config_data['mount_points'] = []

    storage_device = storage_get()
    location = storage_device['location']
    kind = storage_device['kind']
    log('storage attached: {0} at {1}'.format(kind, location))
    if '/' in location:
        entry_id = '_'.join(location.split('/')[-1])
    else:
        entry_id = location

    filesystem = 'ext4'
    success = False
    if not is_mounted(location):
        if create_filesystem(location, filesystem):
            new_mount_point = '/media/{0}_{1}'.format('metadata-storage', entry_id)
            mounted = mount_device(location, new_mount_point, overwrite, filesystem)
            if mounted:
                msg = 'Device {} formatted as {} and mounted as {}.'.format(location,
                                                                            filesystem, new_mount_point)
                status_set('maintenance', msg)
                log(msg)
                config_data['mount_points'].append(new_mount_point)
                success = True, new_mount_point
            else:
                status_set('blocked', 'Mount-point {} exists and is not a folder.'.format(new_mount_point))
                success = False
        else:
            status_set('blocked', 'Could not create filesystem {0} on {1}.'.format(filesystem, location))
            success = False

    if success and is_installed():
        service_restart()
    else:
        # everything done for now
        return 0


@hooks.hook('metadata-storage-storage-detaching')
def metadata_storage_storage_detaching():
    return 0


# -------------------------------------------


@hooks.hook('quobyte-registry-relation-broken')
def qb_reg_rel_broken():
    return 0


@hooks.hook('quobyte-registry-relation-changed')
def qb_reg_rel_changed():
    log('quobyte-registry-relation-changed called', DEBUG)

    return service_restart()


@hooks.hook('quobyte-registry-relation-departed')
def qb_reg_rel_joined():
    return 0


@hooks.hook('quobyte-registry-relation-joined')
def qb_reg_rel_departed():
    return 0


if __name__ == '__main__':
    # execute a hook based on the name the program is called by
    try:
        hooks.execute(sys.argv)
    except UnregisteredHookError as e:
        log('Unknown hook {} - skipping.'.format(e))
