
from charmhelpers.core.hookenv import (
    log,
    ERROR,
    DEBUG,
    config,
    status_set,
    unit_get,
    related_units,
    relation_ids,
    relation_get,
)
from charmhelpers.fetch import (
    apt_install,
    apt_update,
    add_source,
    filter_installed_packages
)
from charmhelpers.core.host import (
    mkdir,
    mount,
    mounts,
    lsb_release
)
from charmhelpers.contrib.network.ip import (
    format_ipv6_addr
)
from tempfile import NamedTemporaryFile
from stat import S_ISBLK
import subprocess
import tempfile
import shutil
import os

# default ports, do not change
api_port_rpc = 7860
registry_port_rpc = 7861
metadata_port_rpc = 7862
data_port_rpc = 7863

web_gui_port = 8080
registry_port_http = 7871


# all default ports
# API: 7860 (TCP), 7865 (TCP, UDP), 7875 (TCP)
# Web Console: 8080 (TCP), 7866 (TCP, UDP), 7876 (TCP)
# Registry Services: 7861 (TCP, UDP), 7871 (TCP)
# Metadata Services: 7862 (TCP, UDP), 7872 (TCP)
# Data Services: 7863 (TCP, UDP), 7873 (TCP)


# -- shell-commands -----------------------------


def call(cmd, level=DEBUG):
    try:
        result = subprocess.check_output(cmd)
    except OSError as e1:
        # cmd not found
        log('command not found: {}'.format(e1), ERROR)
        return False, ''
    except subprocess.CalledProcessError as e2:
        # cmd was not successful
        return False, '{} {}'.format(e2, e2.output)
    log(' '.join(cmd))
    for line in result.split('\n'):
        log(line, level)
    return True, result


def echo_pipe(text, cmd):
    """ wrapper for echo $text | $cmd """
    try:
        subprocess.check_call('echo '+text+' | '+cmd, shell=True)
    except OSError as e:
        log('Error: "echo {} | {}" failed:.\n{}'.format(text, cmd, e), ERROR)
        return False
    return True

# -- file-system --------------------------------


def is_empty(folder):
    return os.listdir(folder) in [[], ['lost+found']]


def is_block_device(filename):
    try:
        mode = os.lstat(filename).st_mode
    except OSError:
        return False
    else:
        return S_ISBLK(mode)


def is_mounted(device):
    # mounts yields [['/mount/point','/dev/path'],[...]]
    return device in map(lambda x, _: x[1], mounts(), [])


def is_quobyte(mount_point):
    return -1 < max([x.upper().find('QUOBYTE_DEV_SETUP') for x in os.listdir(mount_point)])


def create_filesystem(device, filesystem):
    if filesystem in ['ext4']:
        log('Create filesystem {} on {}'.format(filesystem, device))
        success, mkfs_log = call(['mkfs.%s' % filesystem, '-F', '-m0', device])
        if success:
            log('{} filesystem created on {}'.format(filesystem, device))
            return True
        else:
            log(mkfs_log, ERROR)
            return False
    else:
        log('Given parameter filesystem={} is not supported.'.format(filesystem), ERROR)
        return False


def mount_device(device, mount_point, overwrite=True, filesystem="ext4"):
    if os.path.exists(mount_point):
        if os.path.isdir(mount_point):
            mount(device, mount_point, persist=True, filesystem=filesystem)
            return True
        else:
            if overwrite:
                os.unlink(mount_point)
                mount(device, mount_point, persist=True, filesystem=filesystem)
                return True
            else:
                # error ~ mount_point is a file
                return False
    else:
        mkdir(mount_point)
        mount(device, mount_point, persist=True)
        return True


# -- package-system -----------------------------


def fetch_key_from_repository():
    key, keyid = '', ''
    #  fetch key
    release = lsb_release()['DISTRIB_RELEASE']
    repo_id = config('repo_id')
    if release in ['12.04', '12.10', '13.10', '14.04', '14.10', '15.04']:
        url_key = 'https://support.quobyte.com/repo/2/{}/xUbuntu_{}/Release.key'.format(repo_id, release)
    else:
        msg = 'Ubuntu Release {} is not supported'.format(release)
        log(msg, ERROR)
        status_set('blocked', msg)
        return '', ''

    cmd_fetch = ['wget', '-q', '-O', '-', url_key]
    try:
        key = subprocess.check_output(cmd_fetch)
    except subprocess.CalledProcessError as error:
        msg = 'Unable to fetch key: {}'.format(error)
        log(msg, ERROR)
        status_set('blocked', msg)
        return '', ''
    cmd_extract = ['gpg', '--dry-run', '--list-packets']
    p2 = subprocess.Popen(cmd_extract, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    p2.stdin.writelines(key)
    p2.stdin.close()
    key_info = p2.stdout.readlines()
    for line in key_info:
        if line.strip().startswith('keyid'):
            keyid = line.split(':')[1].strip()
            break
    return key, keyid


def update_config_file(key, new_value, filename):
    """ Look for 'key=some_value' in a a given file and replaces some_value with new_value
    :param key: key to lookup
    :param new_value: new value
    :param filename: file to manipulate
    :return: 0 if successful, > 0 otherwise
    """
    try:
        out_file_name = tempfile.mkdtemp()
        out_file_name = os.path.join(out_file_name, os.path.basename(filename))
        out_file = open(out_file_name, 'w')
    except IOError as e:
        log('Cannot open a temp. file: %r' % e, ERROR)
        return 1

    key_found = False
    try:
        with open(filename, 'r') as in_file:
            for line in in_file:
                token = line.strip().split('=')
                if token[0].strip() == key:
                    key_found = True
                    new_line = key+'='+new_value
                    if len(token[1].split('#')):
                        # preserve given comments
                        new_line += ' '+''.join(token[1].split('#')[1:])
                    new_line += '\n'
                else:
                    new_line = line
                out_file.write(new_line)
            if not key_found:
                new_line = '{}={}'.format(key, new_value)
                out_file.write(new_line)
    except IOError as e:
        log('Cannot read from file %r: %r' % (filename, e), ERROR)
        return 1

    out_file.close()
    in_file.close()
    os.remove(filename)
    shutil.move(out_file_name, filename)
    return 0


def is_installed(packages=None):
    if packages is None:
        packages = ['quobyte-client', 'quobyte-server',
                    'oracle-java8-installer', 'ntp']
    return len(filter_installed_packages(packages)) is 0


def install_quobyte():

    if is_installed():
        log('install_quobyte called, but nothing to do', DEBUG)
        return True

    repo_id = config('repo_id')
    oracle_java_license_accepted = config('oracle-java-license-accepted')
    quobyte_license_accepted = config('quobyte-license-accepted')
    if repo_id is None or len(repo_id) == 0:
        status_set('blocked', 'please provide customer repository id using Juju config')
        return False
    if not oracle_java_license_accepted:
        status_set('blocked', 'please accept Oracle Java License using Juju config')
        return False
    if not quobyte_license_accepted:
        status_set('blocked', 'please accept Quobyte License using Juju config')
        return False
    else:
        release = lsb_release()['DISTRIB_RELEASE']
        if release in ['12.04', '14.04', '15.04', '15.10']:
            qb_deb = 'deb https://support.quobyte.com/repo/2/{}/xUbuntu_{} ./'.format(repo_id, release)
            if release in ['12.04']:
                add_source('ppa:ubuntu-toolchain-r/test')
                apt_update(fatal=True)
                apt_install(packages=['python-software-properties'])
        else:
            msg = 'Ubuntu Release {} is not supported'.format(release)
            log(msg, ERROR)
            status_set('blocked', msg)
            return False

    log('Installing Quobyte-Data')
    status_set('maintenance', 'Installing software')

    # install Java
    status_set('maintenance', 'Fetching and installing Java')
    add_source('ppa:webupd8team/java')
    apt_update(fatal=True)
    # accept Oracle Jave License
    if oracle_java_license_accepted:
        echo_pipe('oracle-java8-installer shared/accepted-oracle-license-v1-1 select true',
                  'debconf-set-selections')
    else:
        return False
    apt_install(['oracle-java8-installer'])

    # install Quobyte
    status_set('maintenance', 'Fetching and installing Quobyte')
    key, keyid = fetch_key_from_repository()
    log('add key with keyid "%s" to apt' % keyid)
    add_source(qb_deb, key)
    apt_update(fatal=True)
    log('install packages quobyte-server and ntp')
    apt_install(packages=['quobyte-client', 'quobyte-server', 'ntp'])

    update_config_file('constants.automation.manage_registry_replicas', 'true', r'/etc/quobyte/host.cfg')

    if not is_installed():
        status_set('blocked', 'Installing failed')
        return False

    status_set('maintenance', 'Software is installed')
    return True

# -- Juju-Relations -----------------------------


def get_relation_hosts(relation):
    hosts = []
    address = unit_get('private-address')
    # add self
    hosts.append('{}'.format(format_ipv6_addr(address) or address))
    log('Hosts (local): {}'.format(hosts), DEBUG)
    # add relations
    for rel_id in relation_ids(relation):
        for unit in related_units(rel_id):
            address = relation_get('host', unit, rel_id)
            if address is not None:
                hosts.append('{}'.format(
                    format_ipv6_addr(address) or address))

    hosts.sort()
    log('Hosts (all-known)(relation): {} ({})'.format(hosts, relation), DEBUG)
    return hosts


def do_update_replica_hosts(hosts):
    with_ports = ['{}:{}'.format(host, registry_port_rpc) for host in hosts]
    hosts = ','.join(with_ports)
    success = 0 is update_config_file('registry', hosts, r'/etc/quobyte/host.cfg')
    return success, hosts


# -- quobyte-system -----------------------------


def update_volume_config(key, new_value, configuration_name='BASE'):
    current_file = NamedTemporaryFile()
    new_file = NamedTemporaryFile()
    success, output = call(['qmgmt', 'volume', 'config', 'export', configuration_name, current_file.name])
    if not success:
        log('Error:\n{}'.format(output), ERROR)
    for line in current_file:
        if line.find(':') >= 0:
            current_key, value = line.split(':')
            if key == current_key.strip():
                new_file.write('{}: {}\n'.format(key, new_value))
            else:
                new_file.write(line)
        else:
            new_file.write(line)
    current_file.close()
    new_file.seek(0)
    success, output = call(['qmgmt', 'volume', 'config', 'import', configuration_name, new_file.name])
    if not success:
        log('Error:\n{}'.format(output), ERROR)
    new_file.close()


def fetch_registry_replica_devices(ip_api_service):
    registered_devices = []
    # cmd = ['qmgmt', '--force', '-u',
    cmd = ['qmgmt', '-u',
           'http://{}:{}'.format(ip_api_service, api_port_rpc),
           'registry', 'list']
    success, result = call(cmd)
    if success:
        for line in result:
            if line.strip().isdigit():
                registered_devices.append(line.strip())
        return True, registered_devices
    else:
        log(result, ERROR)
        return False, []


def add_registry_replica(device_id, ip_api_service):
    # cmd = ['qmgmt', '--force', '-u',
    cmd = ['qmgmt', '-u',
           'http://{}:{}'.format(ip_api_service, api_port_rpc),
           'registry', 'add', '{}'.format(device_id)]
    success, result = call(cmd)
    if success:
        log(result)
    else:
        log(result, ERROR)
    return success


def fetch_devices(ip_api_service, service_type=None):
    devices = []
    # cmd = ['qmgmt', '--force', '-u',
    cmd = ['qmgmt', '-u',
           'http://{}:{}'.format(ip_api_service, api_port_rpc),
           'device', 'list']
    success, result = call(cmd)
    if success:
        for line in result:
            token = line.split()
            if (token[0].isdigit() and
                    (service_type is token[-1] or service_type is None)):
                devices.append(token[0])
        return True, devices
    else:
        log(result, ERROR)
        return False, []


def get_device_id(mount_point):
    try:
        with open(os.path.join(mount_point, 'QUOBYTE_DEV_ID'), 'r') as fd:
            for line in fd:
                if line.strip().startswith('device.id'):
                    return line.split('=')[-1].strip()
    except IOError as e:
        log('failed to get quobyte device id from {}: {}'.format(mount_point, e))
        return None
